from groupbuyorganizer import database

from datetime import datetime


class Event(database.Model):
    id = database.Column(database.Integer, primary_key=True)
    name = database.Column(database.String(120), unique=True, nullable=False)
    date_created = database.Column(database.DateTime, nullable=False, default=datetime.utcnow)
    notes = database.Column(database.Text, nullable=True)
    is_locked = database.Column(database.Boolean, nullable=False, default=False)
    is_closed = database.Column(database.Boolean, nullable=False, default=False)
    extra_charges = database.Column(database.Numeric(precision=2), nullable=False, default=0.00)
    added_by = database.Column(database.Integer, database.ForeignKey('user.id'), nullable=False)
    items = database.relationship('Item', backref='event', cascade='all, delete-orphan', lazy='dynamic')


class Item(database.Model):
    id = database.Column(database.Integer, primary_key=True)
    name = database.Column(database.Integer, nullable=False)
    price = database.Column(database.Numeric(precision=2), nullable=False)
    packing = database.Column(database.Integer, nullable=False)
    category_id = database.Column(database.Integer, database.ForeignKey('category.id'), nullable=False)
    event_id = database.Column(database.Integer, database.ForeignKey('event.id'), nullable=False)
    added_by = database.Column(database.Integer, database.ForeignKey('user.id'), nullable=False)
    case_buys = database.relationship('CaseBuy', backref='item', cascade='all, delete-orphan', lazy='dynamic')
    case_splits = database.relationship('CaseSplit', backref='item', cascade='all, delete-orphan', lazy='dynamic')

    def __repr__(self):
        return f"{self.name}"


class CaseBuy(database.Model):
    id = database.Column(database.Integer, primary_key=True)
    user_id = database.Column(database.Integer, database.ForeignKey('user.id'), nullable=False)
    event_id = database.Column(database.Integer, database.ForeignKey('event.id'), nullable=False)
    item_id = database.Column(database.Integer, database.ForeignKey('item.id'), nullable=False)
    quantity = database.Column(database.Integer, nullable=False)


class CaseSplit(database.Model):
    id = database.Column(database.Integer, primary_key=True)
    started_by = database.Column(database.Integer, database.ForeignKey('user.id'), nullable=False)
    date_created = database.Column(database.DateTime, nullable=False, default=datetime.utcnow)
    event_id = database.Column(database.Integer, database.ForeignKey('event.id'), nullable=False)
    item_id = database.Column(database.Integer, database.ForeignKey('item.id'), nullable=False)
    is_complete = database.Column(database.Boolean, nullable=False, default=False)
    commits = database.relationship('CasePieceCommit', backref='case_split', cascade='all, delete-orphan')


class CasePieceCommit(database.Model):
    id = database.Column(database.Integer, primary_key=True)
    committed_on = database.Column(database.DateTime, nullable=False, default=datetime.utcnow)
    event_id = database.Column(database.Integer, database.ForeignKey('event.id'), nullable=False)
    case_split_id = database.Column(database.Integer, database.ForeignKey('case_split.id'), nullable=False)
    user_id = database.Column(database.Integer, database.ForeignKey('user.id'), nullable=False)
    item_id = database.Column(database.Integer, database.ForeignKey('item.id'), nullable=False)
    pieces_committed = database.Column(database.Integer, nullable=False)


# class HasPaid(database.Model):
#     user_id = database.Column(database.Integer, database.ForeignKey('item.id'), nullable=False)
#     event_id = database.Column(database.Integer, database.ForeignKey('event.id'), nullable=False)
#     has_paid = database.Column(database.Boolean, nullable=False, default=False)


from flask_login import current_user
from flask_wtf import FlaskForm
from wtforms import BooleanField, PasswordField, StringField, SubmitField
from wtforms.validators import DataRequired, Email, EqualTo, Length, ValidationError

from groupbuyorganizer.admin.models import User


class RegistrationForm(FlaskForm):
    '''This field is used to register new accounts'''

    username = StringField('Username', validators=[DataRequired(), Length(min=2, max=20)])
    email = StringField('Email', validators=[DataRequired(), Email()])
    password = PasswordField('Password', validators=[DataRequired(), Length(min=6)])
    confirm_password = PasswordField('Confirm Password', validators=[DataRequired(), EqualTo('password')])
    submit = SubmitField('Sign Up')

    def validate_username(self, username):
        user = User.query.filter_by(username=username.data).first()
        if user:
            raise ValidationError('That username is taken. Please choose a different one.')

    def validate_email(self, email):
        user = User.query.filter_by(email=email.data).first()
        if user:
            raise ValidationError('That email is taken. Please choose a different one.')


class LoginForm(FlaskForm):
    '''This form is used to log into the web app.'''
    email = StringField('Email', validators=[DataRequired(), Email()])
    password = PasswordField('Password', validators=[DataRequired()])
    remember = BooleanField('Remember Me')
    submit = SubmitField('Log In')


class UserOptionsForm(FlaskForm):
    '''This field is used to register new accounts'''

    username = StringField('Username (displayed to other users)', validators=[DataRequired(), Length(min=2, max=20)])
    email = StringField('Email (used for logging in)', validators=[DataRequired(), Email()])
    submit = SubmitField('Update')

    def validate_username(self, username):
        if username.data != current_user.username:
            user = User.query.filter_by(username=username.data).first()
            if user:
                raise ValidationError('That username is taken. Please choose a different one.')

    def validate_email(self, email):
        if email.data != current_user.email:
            print(f'{email.data} {current_user.username}')
            user = User.query.filter_by(email=email.data).first()
            if user:
                raise ValidationError('That email is taken. Please choose a different one.')


from flask import Blueprint, flash, make_response, redirect, render_template, request, url_for
from flask_login import current_user, login_required
import pdfkit

from groupbuyorganizer import database
from groupbuyorganizer.admin.models import Category, Instance, User
from groupbuyorganizer.events.forms import CreateItemForm, CreateCaseSplitForm, CreateEventForm, CaseQuantityOrderForm,\
    EditItemForm, EventExtraChargeForm, SelectUserFromEventForm
from groupbuyorganizer.events.models import CaseBuy, CasePieceCommit, CaseSplit, Event, Item
from groupbuyorganizer.events.utility_functions import fetch_user_items, generate_active_user_select_field,\
    is_event_active, is_user_active, return_qty_price_select_field
from groupbuyorganizer.events.utility_objects import BreakdownObject, CaseSplitItemGroup, EventItem, \
    MyOrderObject, SummaryObject,  StructuredEventItemList, UserTotalItem


events = Blueprint('events', __name__)

@events.route('/events/<int:event_id>/items/', methods=['GET', 'POST'])
@events.route('/events/<int:event_id>/', methods=['GET', 'POST'])
@login_required
def event(event_id):

    # Item add form setup
    instance = Instance.query.first()
    event = Event.query.get_or_404(event_id)
    available_categories = Category.query.order_by('name')
    categories_list = [(piece.id, piece.name) for piece in available_categories]
    form = CreateItemForm()
    form.category_id.choices = categories_list

    # Remove user from event form
    remove_user_from_event_form = SelectUserFromEventForm()
    remove_user_from_event_form.user_to_select.choices = generate_active_user_select_field(event.id)

    #Items Setup
    items = database.session.query(Item, Category.name).filter_by(event_id=event.id).join(Category, Item.category_id
                                                            == Category.id).order_by(Category.name, Item.name).all()
    structured_item_list = None
    if items:
        structured_item_list = StructuredEventItemList(items, event_id)

    if form.validate_on_submit() and form.item_name.data:
        item = Item(name=form.item_name.data, price=form.price.data, packing=form.packing.data,
                    category_id=form.category_id.data, added_by=current_user.get_id(), event_id=event_id)
        database.session.add(item)
        database.session.commit()
        flash('Item successfully added!', 'success')
        return redirect(url_for('events.event', event_id=event_id))

    if remove_user_from_event_form.validate_on_submit():
        if current_user.is_admin == False:
            flash('Access denied', 'danger')
            return redirect(url_for('general.home'))
        active_user = User.query.filter_by(id=remove_user_from_event_form.user_to_select.data).first()
        case_buys = CaseBuy.query.filter_by(event_id=event_id, user_id=active_user.id).all()
        for case_buy in case_buys:
            database.session.delete(case_buy)
        case_piece_commits = CasePieceCommit.query.filter_by(user_id=active_user.id, event_id=event_id).all()
        case_piece_commit_ids = []
        for case_piece_commit in case_piece_commits:
            case_piece_commit_ids.append(case_piece_commit.id)
            database.session.delete(case_piece_commit)
            database.session.commit()
        affected_case_splits = database.session.query(CaseSplit).filter(CaseSplit.id.in_(case_piece_commit_ids)).all()

        # Removing case splits when user was only partipants, and changing the rest to is_complete=False if they were
        # initially marked as complete.
        for case_split in affected_case_splits:
            if not case_split.commits:
                database.session.delete(case_split)
            else:
                if case_split.is_complete == True:
                    case_split.is_complete = False
        database.session.commit()

        database.session.commit()
        flash(f'All transactions for {active_user.username} successfully removed from event!', 'info')
        return redirect(url_for('events.event', event_id=event_id))

    return render_template('event.html', event=event, form=form, title=f'{event.name} Overview',
                           structured_item_list=structured_item_list,
                           remove_user_from_event_form=remove_user_from_event_form,
                           users_can_see_master_overview=instance.users_can_see_master_overview)


@events.route('/events/<int:event_id>/edit', methods=['GET', 'POST'])
@login_required
def event_edit(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    event_name_form = CreateEventForm()
    event_edit_form = EventExtraChargeForm()
    if event_name_form.validate_on_submit():
        event.name = event_name_form.event_name.data
        database.session.commit()
        flash(f'{event.name} has been edited!', 'info')
        return redirect(url_for('events.event_edit', event_id=event_id))
    if event_edit_form.validate_on_submit():
        event.extra_charges = event_edit_form.extra_charges.data
        event.notes = event_edit_form.event_notes.data
        database.session.commit()
        flash('Extra charges updated!', 'info')
        return redirect(url_for('events.event_edit', event_id=event_id))
    elif request.method == 'GET':
        event_name_form.event_name.data = event.name
        event_edit_form.extra_charges.data = event.extra_charges
        if event.notes:
            event_edit_form.event_notes.data = event.notes

    return render_template('event_edit.html', title=f'{event.name} - Edit', event_name_form=event_name_form,
                           event_extra_charge_form=event_edit_form, event=event)


@events.route("/category_settings/<int:event_id>/remove/", methods=['GET'])
@login_required
def event_remove(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    database.session.delete(event)
    database.session.commit()
    flash('Event deleted!', 'info')
    return redirect(url_for('general.home'))


@events.route("/category_settings/<int:event_id>/lock/", methods=['GET'])
@login_required
def event_lock(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    event.is_locked = True
    database.session.commit()
    flash('Event locked!', 'info')
    return redirect(url_for('events.event_edit', event_id=event_id))


@events.route("/category_settings/<int:event_id>/unlock/", methods=['GET'])
@login_required
def event_unlock(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    event.is_locked = False
    database.session.commit()
    flash('Event unlocked!', 'info')
    return redirect(url_for('events.event_edit', event_id=event_id))


@events.route("/category_settings/<int:event_id>/close/", methods=['GET'])
@login_required
def event_close(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    event.is_locked = True
    event.is_closed = True
    database.session.commit()
    flash('Event closed!', 'info')
    return redirect(url_for('events.event_edit', event_id=event_id))


@events.route("/category_settings/<int:event_id>/open/", methods=['GET'])
@login_required
def event_open(event_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    event.is_closed = False
    database.session.commit()
    flash('Event opened!', 'info')
    return redirect(url_for('events.event_edit', event_id=event_id))


@events.route('/events/<int:event_id>/items/<int:item_id>/', methods=['GET', 'POST'])
@login_required
def item(event_id, item_id):
    event = Event.query.get_or_404(event_id)
    item = Item.query.get_or_404(item_id)

    # Item edit form
    available_categories = Category.query.order_by('name')
    categories_list = [(piece.id, piece.name) for piece in available_categories]
    edit_item_form = EditItemForm()
    edit_item_form.category_id.choices = categories_list

    # Added by name
    added_by_user = User.query.filter_by(id=item.added_by).first()

    # Case Quantity Form
    order_case_form = CaseQuantityOrderForm()
    order_case_form.quantity.choices = return_qty_price_select_field(100, item.price, item.packing,
                                                                     whole_cases=True)

    # Create Case Splits
    create_case_split_form = CreateCaseSplitForm()
    create_case_split_form.piece_quantity.choices = \
        return_qty_price_select_field(item.packing - 1, item.price, item.packing)

    # Customized item/order view
    event_item = EventItem(item, event_id)

    if edit_item_form.validate_on_submit():
        item.name = edit_item_form.item_name.data
        item.category_id = edit_item_form.category_id.data
        item.price = edit_item_form.price.data
        additional_flash_message = ""
        if item.packing != edit_item_form.packing.data:
            additional_flash_message = 'Because of the change in item packing, all case splits have been removed.'
            case_splits = CaseSplit.query.filter_by(item_id=item.id)
            for case_split in case_splits:
                database.session.delete(case_split)
        item.packing = edit_item_form.packing.data
        database.session.commit()
        flash(f'Item successfully edited!  {additional_flash_message}', 'info')
        return redirect(url_for('events.event', event_id=event_id))

    elif order_case_form.validate_on_submit():
        previous_order = CaseBuy.query.filter_by(user_id=current_user.id, event_id=event_id, item_id=item.id).first()
        if not previous_order:
            case_buy = CaseBuy(user_id=current_user.id, event_id=event_id, item_id=item_id,
                               quantity=order_case_form.quantity.data)
            database.session.add(case_buy)
            database.session.commit()
            flash('Case order for item added!', 'success')
            return redirect(url_for('events.item', event_id=event_id, item_id=item.id))
        else:
            if order_case_form.quantity.data == 0:
                database.session.delete(previous_order)
                database.session.commit()
                flash('Case(s) removed from your order!', 'info')
                return redirect(url_for('events.item', event_id=event_id, item_id=item.id))
            else:
                previous_order.quantity = order_case_form.quantity.data
                database.session.commit()
                flash('Case quantity updated!', 'info')
                return redirect(url_for('events.item', event_id=event_id, item_id=item.id))

    elif create_case_split_form.validate_on_submit():
        case_split = CaseSplit(started_by=current_user.id, event_id=event.id, item_id=item.id, is_complete=False)
        database.session.add(case_split)
        database.session.commit()
        case_piece_commit = CasePieceCommit(case_split_id=case_split.id, user_id=current_user.id, event_id=event_id,
                                            pieces_committed=create_case_split_form.piece_quantity.data,
                                            item_id=item.id)
        database.session.add(case_piece_commit)
        database.session.commit()
        flash('Case split created!', 'success')
        return redirect(url_for('events.item', event_id=event.id, item_id=item.id))

    elif request.method == 'GET':
        edit_item_form.category_id.choices = categories_list
        edit_item_form.item_name.data = item.name
        edit_item_form.category_id.data = item.category_id
        edit_item_form.price.data = item.price
        edit_item_form.packing.data = item.packing

        #populating current case buy quantity, if any
        previous_order = CaseBuy.query.filter_by(user_id=current_user.id, event_id=event.id, item_id=item.id).first()
        if previous_order:
            order_case_form.quantity.data = previous_order.quantity


    return render_template('item.html', added_by_user=added_by_user, form=edit_item_form, item_id=item.id,
                           order_case_form=order_case_form, event=event, item=event_item,
                           create_case_split_form=create_case_split_form, title=f'{item.name} Overview')


@events.route('/events/<int:event_id>/items/<int:item_id>/remove/', methods=['GET'])
@login_required
def remove_item(event_id, item_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    event = Event.query.get_or_404(event_id)
    item = Item.query.get_or_404(item_id)
    database.session.delete(item)
    database.session.commit()
    flash('Item successfully removed!', 'info')
    return redirect(url_for('events.event', event_id=event_id))


@events.route('/events/<int:event_id>/event_summary/')
@login_required
def event_summary(event_id):
    event = Event.query.get_or_404(event_id)
    instance = Instance.query.first()
    if instance.users_can_see_master_overview == False:
        if current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    event_active = is_event_active(event.id)
    order_object = None
    if event_active == True:
        order_object = SummaryObject(event)

    case_buy_items = database.session.query(Item, Category.name).filter(Item.event_id == event.id,
                                                                        Item.id == CaseBuy.item_id)\
                                                        .join(Category, Item.category_id == Category.id)\
                                                        .order_by(Category.name, Item.name).all()


    return render_template('event_summary.html', event=event, is_pdf=False, order_object=order_object,
                           users_can_see_master_overview=instance.users_can_see_master_overview, title=f'{event.name} '
        f'Summary')


@events.route('/events/<int:event_id>/event_summary/pdf')
@login_required
def event_summary_pdf(event_id):
    event = Event.query.get_or_404(event_id)
    instance = Instance.query.first()
    if instance.users_can_see_master_overview == False:
        if current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    event_active = is_event_active(event.id)
    order_object = None
    if event_active == True:
        order_object = SummaryObject(event)

    # PDF Setup
    config = pdfkit.configuration(wkhtmltopdf=instance.wkhtmltopdf_path)
    rendered = render_template('event_summary.html', is_pdf=True, event=event, order_object=order_object,
                               title=f'{event.name} Order Overview')
    pdf = pdfkit.from_string(rendered, False, configuration=config, options={'quiet': ''})
    response = make_response(pdf)
    response.headers['Content-Type'] = 'application/pdf'
    response.headers['Content-Disposition'] = f'attachment; filename=Order_Summary.pdf'
    return response


@events.route('/events/<int:event_id>/event_total_user_breakdown/', methods=['GET', 'POST'])
@login_required
def event_total_user_breakdown(event_id):
    event = Event.query.get_or_404(event_id)
    instance = Instance.query.first()
    if instance.users_can_see_master_overview == False:
        if current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    event_active = is_event_active(event.id)
    order_object = None
    if event_active == True:
        order_object = BreakdownObject(event)

    user_total_object = UserTotalItem(event)

    return render_template('user_breakdown.html', event=event, is_pdf=False, breakdown_object=order_object,
                           users_can_see_master_overview=instance.users_can_see_master_overview,
                           user_total_object=user_total_object, title=f'{event.name} Event Total -- User Breakdown')


@events.route('/events/<int:event_id>/event_total_user_breakdown/pdf/', methods=['GET'])
@login_required
def event_total_user_breakdown_pdf(event_id):
    event = Event.query.get_or_404(event_id)
    instance = Instance.query.first()
    if instance.users_can_see_master_overview == False:
        if current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    event_active = is_event_active(event.id)
    order_object = None
    if event_active == True:
        order_object = BreakdownObject(event)

    user_total_object = UserTotalItem(event)

    # PDF Setup
    config = pdfkit.configuration(wkhtmltopdf=instance.wkhtmltopdf_path)
    rendered = render_template('user_breakdown.html', is_pdf=True, event=event, user_total_object=user_total_object,
                               title=f'{event.name} - Case Breakdown', breakdown_object=order_object)
    pdf = pdfkit.from_string(rendered, False, configuration=config, options={'quiet': ''})
    response = make_response(pdf)
    response.headers['Content-Type'] = 'application/pdf'
    response.headers['Content-Disposition'] = f'attachment; filename=Event_Overview_User_Breakdown.pdf'
    return response


@events.route('/events/<int:event_id>/user_order/<int:user_id>/', methods=['GET', 'POST'])
@login_required
def my_order(event_id, user_id):
    event = Event.query.get_or_404(event_id)
    user = User.query.get_or_404(user_id)
    instance = Instance.query.first()

    if user_id != current_user.id:
        if instance.users_can_see_master_overview == False and current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    form = SelectUserFromEventForm()
    form.user_to_select.choices = generate_active_user_select_field(event.id)

    user_active = is_user_active(user.id, event.id)
    order_object = None
    if user_active == True:
        user_items = fetch_user_items(user.id, event.id)
        order_object = MyOrderObject(user, user_items, event)

    if form.validate_on_submit():
        user_to_redirect_to = form.user_to_select.data
        return redirect(url_for('events.my_order', event_id=event.id, user_id=user_to_redirect_to))

    if request.method == 'GET':
        form.user_to_select.data = user.id



    return render_template('my_order.html', event=event, is_pdf=False, user_name = user.username, user_id=user.id,
                           user_active=user_active, order_object=order_object, form=form,
                           users_can_see_master_overview=instance.users_can_see_master_overview,
                           title=f"{user.username}'s order")


@events.route('/events/<int:event_id>/user_order/<int:user_id>/pdf/', methods=['GET', 'POST'])
@login_required
def my_order_pdf(event_id, user_id):
    event = Event.query.get_or_404(event_id)
    user = User.query.get_or_404(user_id)
    instance = Instance.query.first()

    if instance.users_can_see_master_overview == False and current_user.id != user.id:
        if current_user.is_admin == False:
            flash('Access denied', 'warning')
            return redirect(url_for('general.home'))

    user_active = is_user_active(user.id, event.id)
    order_object = None
    if user_active == True:
        user_items = fetch_user_items(user.id, event.id)
        order_object = MyOrderObject(user, user_items, event)

    # PDF Setup
    config = pdfkit.configuration(wkhtmltopdf=instance.wkhtmltopdf_path)
    rendered = render_template('my_order.html', is_pdf=True, event=event, user_active=user_active, user_id=user.id,
                               form=None, order_object=order_object, user_name = user.username,
                               title=f"{user.username}'s order")
    pdf = pdfkit.from_string(rendered, False, configuration=config, options={'quiet': ''})
    response = make_response(pdf)
    response.headers['Content-Type'] = 'application/pdf'
    response.headers['Content-Disposition'] = f'attachment; filename=Order_Overview.pdf'
    return response


@events.route('/events/<int:event_id>/items/<int:item_id>/case_splits/<int:case_split_id>/remove/', methods=['GET'])
@login_required
def remove_case_split(event_id, item_id, case_split_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))

    event = Event.query.get_or_404(event_id)
    item = Item.query.get_or_404(item_id)
    case_split = CaseSplit.query.get_or_404(case_split_id)
    database.session.delete(case_split)
    database.session.commit()
    flash('Case split removed!', 'info')
    return redirect(url_for('events.item', event_id=event.id, item_id=item.id))


@events.route('/events/<int:event_id>/items/<int:item_id>/case_splits/<int:case_split_id>/commits/<int:commit_id>/remove',
    methods=['GET'])
@login_required
def remove_case_split_pledge(event_id, item_id, case_split_id, commit_id):
    event = Event.query.get_or_404(event_id)
    item = Item.query.get_or_404(item_id)
    case_split = CaseSplit.query.get_or_404(case_split_id)
    commit = CasePieceCommit.query.get_or_404(commit_id)
    if commit.user_id == current_user.id or current_user.is_admin:
        database.session.delete(commit)
        if case_split.is_complete == True:
            case_split.is_complete = False
        database.session.commit()
        if not case_split.commits:
            database.session.delete(case_split)
            database.session.commit()
        flash('Case split commit removed!', 'info')
        return redirect(url_for('events.item', event_id=event.id, item_id=item.id))
    else:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))


@events.route('/events/<int:event_id>/items/<int:item_id>/case_splits/<int:case_split_id>/',
    methods=['GET', 'POST'])
@login_required
def case_split(event_id, item_id, case_split_id):
    event = Event.query.get_or_404(event_id)
    item = Item.query.get_or_404(item_id)
    case_split = CaseSplit.query.get_or_404(case_split_id)
    case_split_group = CaseSplitItemGroup((case_split,), item.packing, event.id, is_single=True)
    created_by = case_split_group.single_split[1][0]
    case_split_item = case_split_group.single_split[0]

    modify_case_split_form = CreateCaseSplitForm()
    pieces_reserved_so_far = 0

    # This is what prevents a single user from being the sole user of a case split, forcing him to just buy a
    # case.  Aside from that, this doesn't include the user itself in the count for the max items to pledge.
    def how_many_pieces_reserved(commits, current_user_id):
        '''Short function used twice to calculate how many items are already accounted for in this case split.'''
        pieces_reserved_so_far = 0
        if len(commits) == 1 and commits[0].user_id == current_user_id:
            pieces_reserved_so_far = 1
        else:
            for commit in case_split.commits:
                if commit.user_id != current_user_id:
                    pieces_reserved_so_far += commit.pieces_committed
        return pieces_reserved_so_far

    # Evaluate case split is_complete status
    def check_is_complete(case_split):
        piece_sum = 0
        for commit in case_split.commits:
            piece_sum += commit.pieces_committed
        if case_split.is_complete:
            if piece_sum < item.packing:
                case_split.is_complete = False
        else:
            if piece_sum == item.packing:
                case_split.is_complete = True
        database.session.commit()

    pieces_reserved_so_far = how_many_pieces_reserved(case_split.commits, current_user.id)

    modify_case_split_form.piece_quantity.choices = return_qty_price_select_field(item.packing - pieces_reserved_so_far,
                                                                                  item.price, item.packing)
    case_piece_commit = CasePieceCommit.query.filter_by(case_split_id=case_split.id,
                                                        user_id=current_user.id).first()

    if modify_case_split_form.validate_on_submit():
        case_split = CaseSplit.query.filter_by(id=case_split.id).first()
        if case_split.is_complete:
            flash('This case split has closed in between you loading the previous page and clicking submit... '
                  ' someone else beat you to it!  Start a new case split instead.', 'warning')
            return redirect(url_for('events.item', event_id=event.id, item_id=item.id))
        else:
            case_piece_commit = CasePieceCommit.query.filter_by(case_split_id=case_split.id, user_id=current_user.id).first()
            if case_piece_commit:
                # Validating there are enough pieces left.  Since javascript isn't implemented in this library (yet), the
                # number of pieces in a case split can change without any warning to the user.
                case_split = CaseSplit.query.filter_by(id=case_split_id).first()
                if case_split:
                    pieces_reserved_so_far = how_many_pieces_reserved(case_split.commits, current_user.id)
                    if modify_case_split_form.piece_quantity.data + pieces_reserved_so_far > item.packing:
                        flash(
                            'Items pledged exceeds what is currently available for this case split.  This occurs when someone'
                            ' else places a pledge right before you.  Please resubmit a pledge with a different quantity, or '
                            'otherwise create a new case split.', 'warning')
                        return redirect(url_for('events.case_split', event_id=event.id, item_id=item.id,
                                                case_split_id=case_split.id))
                    else:
                        case_piece_commit.pieces_committed = modify_case_split_form.piece_quantity.data
                        database.session.commit()
                        check_is_complete(case_split)
                        flash(
                            'Case split pledge updated!', 'info')
                        return redirect(url_for('events.case_split', event_id=event.id, item_id=item.id,
                                                case_split_id=case_split.id))
            else:
                pieces_reserved_so_far = how_many_pieces_reserved(case_split.commits, current_user.id)
                if modify_case_split_form.piece_quantity.data + pieces_reserved_so_far > item.packing:
                    flash(
                        'Items pledged exceeds what is currently available for this case split.  This occurs when '
                        'someone else places a pledge right before you.  Please resubmit a pledge with a different '
                        'quantity, or otherwise create a new case split.', 'warning')
                    return redirect(url_for('events.case_split', event_id=event.id, item_id=item.id,
                                            case_split_id=case_split.id))
                else:
                    commit = CasePieceCommit(event_id=event.id, case_split_id=case_split.id, user_id=current_user.id,
                                                        item_id=item.id,
                                                         pieces_committed=modify_case_split_form.piece_quantity.data)
                    database.session.add(commit)
                    database.session.commit()
                    check_is_complete(case_split)
                    flash(
                        'Case split pledge created!', 'info')
                    return redirect(url_for('events.case_split', event_id=event.id, item_id=item.id,
                                            case_split_id=case_split.id))

    elif request.method == 'GET':
        if case_piece_commit:
            modify_case_split_form.piece_quantity.data = case_piece_commit.pieces_committed
    return render_template('case_split.html', event=event, item=item, created_by=created_by,
                           form=modify_case_split_form, case_split=case_split_item, is_locked=event.is_locked,
                           title=f'{item.name} Case Split')


from flask_login import UserMixin

from datetime import datetime

from groupbuyorganizer import database, login_manager


@login_manager.user_loader
def load_user(user_id):
    return User.query.get(int(user_id))


class User(database.Model, UserMixin):

    id = database.Column(database.Integer, primary_key=True)
    username = database.Column(database.String(20), unique=True, nullable=False)
    password = database.Column(database.String(60), nullable=False)
    email = database.Column(database.String(120), unique=True, nullable=False)
    disabled = database.Column(database.Boolean, nullable=False, default=False)
    is_admin = database.Column(database.Boolean, nullable=False, default=False)
    is_root = database.Column(database.Boolean, nullable=False, default=False)
    date_created = database.Column(database.DateTime, nullable=False, default=datetime.utcnow)


    def __repr__(self):
        return f"User('{self.username}', '{self.email}')"


class Instance(database.Model):

    id = database.Column(database.Integer, primary_key=True)
    root_created = database.Column(database.Boolean, nullable=False, default=False)
    registration_enabled = database.Column(database.Boolean, nullable=False, default=True)
    users_can_see_master_overview = database.Column(database.Boolean, nullable=False, default=True)
    wkhtmltopdf_path = database.Column(database.String(256), nullable=True)


class Category(database.Model):

    id = database.Column(database.Integer, primary_key=True)
    name = database.Column(database.String(100), nullable=False, unique=True)
    items = database.relationship('Item', backref='category', lazy=True)


# -*- coding: utf-8 -*-
"""
    flask_security.babel
    ~~~~~~~~~~~~~~~~~~~~

    I18N support for Flask-Security.
"""

from flask_babelex import Domain
from wtforms.i18n import messages_path

wtforms_domain = Domain(messages_path(), domain='wtforms')


class Translations(object):
    """Fixes WTForms translation support and uses wtforms translations."""

    def gettext(self, string):
        return wtforms_domain.gettext(string)

    def ngettext(self, singular, plural, n):
        return wtforms_domain.ngettext(singular, plural, n)



from flask import Blueprint, flash, redirect, render_template, request, url_for
from flask_login import current_user, login_required, login_user, logout_user

from groupbuyorganizer import database, bcrypt
from groupbuyorganizer.admin.models import Instance, User
from groupbuyorganizer.admin.utilities import HomeEvent
from groupbuyorganizer.events.forms import CreateEventForm
from groupbuyorganizer.events.models import Event
from groupbuyorganizer.general.forms import LoginForm, RegistrationForm, UserOptionsForm


general = Blueprint('general', __name__)

@general.route("/events/")
@general.route("/", methods=['GET', 'POST'])
def home():
    form = CreateEventForm()
    instance = Instance.query.first()
    events = Event.query.order_by(Event.date_created.desc()).all()

    home_event_list = []
    for event in events:
        home_event_list.append(HomeEvent(event))

    if form.validate_on_submit():
        event = Event(name=form.event_name.data, added_by=current_user.id)
        database.session.add(event)
        database.session.commit()
        flash('Event created!', 'success')
        return redirect(url_for('general.home'))
    return render_template('home.html', root_created = instance.root_created, home_event_list=home_event_list,
                           registration_enabled = instance.registration_enabled, events=events, form=form,
                           users_can_see_master_overview=instance.users_can_see_master_overview)


@general.route("/about/")
def about():
    return render_template('about.html', title='About')


@general.route("/register/", methods=['GET', 'POST'])
def register():
    if current_user.is_authenticated:
        flash('You are already logged in!', 'info')
        return redirect(url_for('general.home'))
    instance = Instance.query.first()
    form = RegistrationForm()
    if form.validate_on_submit():
        hashed_password = bcrypt.generate_password_hash(form.password.data).decode('utf-8')
        user = User(username=form.username.data, email=form.email.data, password=hashed_password)
        instance = Instance.query.first()
        if instance.root_created == False:
            user.is_root = True
            user.is_admin = True
            instance.root_created = True
            flash('Your account has been created!', 'success')
        else:
            flash(f'Your account has been created, you can now log in', 'success')

        database.session.add(user)
        database.session.commit()
        return redirect(url_for('general.login'))
    return render_template('register.html', title='Join Today', form=form,
                           registration_enabled=instance.registration_enabled)


@general.route("/login/", methods=['GET', 'POST'])
def login():
    instance = Instance.query.first()
    if current_user.is_authenticated:
        return redirect(url_for('general.home'))
    form = LoginForm()
    if form.validate_on_submit():
        user = User.query.filter_by(email=form.email.data).first()
        if user and bcrypt.check_password_hash(user.password, form.password.data):
            if user.disabled:
                flash('This account has been disabled.', 'danger')
                return redirect(url_for('general.home'))
            login_user(user, remember=form.remember.data)
            next_page = request.args.get('next')
            return redirect(next_page) if next_page else redirect(url_for('general.home'))
        else:
            flash('Login Unsuccessful. Please check email and password', 'danger')
    return render_template('login.html', title='Log In', form=form, registration_enabled=instance.registration_enabled)


@general.route("/logout/")
def logout():
    logout_user()
    return redirect(url_for('general.home'))


@general.route("/account/", methods=['GET', 'POST'])
@login_required
def account():
    form = UserOptionsForm()
    if form.validate_on_submit():
        current_user.username = form.username.data
        current_user.email = form.email.data
        database.session.commit()
        flash('Your account has been updated!', 'success')
        return redirect(url_for('general.account'))
    elif request.method == 'GET':
        form.username.data = current_user.username
        form.email.data = current_user.email
    return render_template('account.html', title=f'{current_user.username} Account Settings', form=form)


@general.route("/help/")
def help():
    return render_template('help.html', title='Help')


from flask import flash
from flask_login import current_user

from groupbuyorganizer import database
from groupbuyorganizer.admin.models import User
from groupbuyorganizer.events.models import CaseSplit, Item, Event
from groupbuyorganizer.events.utility_objects import get_active_participants, get_case_list


def admin_protector(user):
    '''This function protects against those with admin power from manually inputting valid route commands to user
    functions, even when the buttons are disabled.
    '''

    if user.is_admin and current_user.is_admin and current_user.is_root == False:
        flash('Access denied.', 'danger')


class HomeEvent:
    def __init__(self, event):
        self.event = event
        self.case_list = get_case_list(self.event.id)
        self.added_by = self.get_added_by()
        self.active_participants = get_active_participants(self.event.id, return_length=True)
        self.active_case_splits = self.get_active_case_splits()
        self.total_cases = self.get_total_cases()
        self.event_total = self.get_event_total()

    def get_added_by(self):
        user = User.query.filter_by(id=Event.added_by).first()
        return user.username


    def get_active_case_splits(self):
        active_case_splits = database.session.query(CaseSplit.id).filter(CaseSplit.event_id == self.event.id,
                                                        CaseSplit.is_complete == False).all()
        return len(active_case_splits)

    def get_total_cases(self):
        total_count = 0
        for case_order in self.case_list[0]:
            total_count += case_order.quantity
        for case_split in self.case_list[1]:
            if case_split.is_complete == True:
                total_count += 1
        return total_count

    def get_event_total(self):
        total_cost = 0
        for case_order in self.case_list[0]:
            item = Item.query.filter_by(id=case_order.item_id).first()
            total_cost += (case_order.quantity * item.price)

        for case_split in self.case_list[1]:
            if case_split.is_complete == True:
                item = Item.query.filter_by(id=case_split.item_id).first()
                total_cost += item.price
        return total_cost


# MIT License
#
# Copyright (c) 2019 - ∞ Mark Michon
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in all
# copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
# SOFTWARE.

# Project page:
# https://github.com/MarkMichon1/Group-Buy-Organizer

# Contact:
# markmichon7@gmail.com

from flask import Flask
from flask_bcrypt import Bcrypt
from flask_login import LoginManager
from flask_sqlalchemy import SQLAlchemy

import binascii
import os


web_app = Flask(__name__)
web_app.config['SECRET_KEY'] = binascii.b2a_hex(os.urandom(32))
web_app.config['SQLALCHEMY_DATABASE_URI'] = 'sqlite:///site.db'

database = SQLAlchemy(web_app)
bcrypt = Bcrypt(web_app)
login_manager = LoginManager(web_app)
login_manager.login_view = 'general.login'
login_manager.login_message_category = 'general.info'

from groupbuyorganizer.admin.models import Category, Instance
import groupbuyorganizer.events.models

from groupbuyorganizer.general.routes import general
from groupbuyorganizer.events.routes import events
from groupbuyorganizer.admin.routes import admin
from groupbuyorganizer.errors.handlers import errors
web_app.register_blueprint(general)
web_app.register_blueprint(events)
web_app.register_blueprint(admin)
web_app.register_blueprint(errors)

database.create_all()
database.session.commit()

# Creating an "Instance" model if there is none.
if Instance.query.get(1) is None:
    database.session.add(Instance(wkhtmltopdf_path="C:\\Program Files\\wkhtmltopdf\\bin\\wkhtmltopdf.exe"))
    database.session.add(Category(name='Uncategorized'))
    database.session.commit()


# -*- coding: utf-8 -*-
"""
    flask_security.changeable
    ~~~~~~~~~~~~~~~~~~~~~~~~~

    Flask-Security recoverable module

    :copyright: (c) 2012 by Matt Wright.
    :author: Eskil Heyn Olsen
    :license: MIT, see LICENSE for more details.
"""

from flask import current_app
from werkzeug.local import LocalProxy

from .signals import password_changed
from .utils import config_value, hash_password, send_mail

# Convenient references
_security = LocalProxy(lambda: current_app.extensions['security'])

_datastore = LocalProxy(lambda: _security.datastore)


def send_password_changed_notice(user):
    """Sends the password changed notice email for the specified user.

    :param user: The user to send the notice to
    """
    if config_value('SEND_PASSWORD_CHANGE_EMAIL'):
        subject = config_value('EMAIL_SUBJECT_PASSWORD_CHANGE_NOTICE')
        send_mail(subject, user.email, 'change_notice', user=user)


def change_user_password(user, password):
    """Change the specified user's password

    :param user: The user to change_password
    :param password: The unhashed new password
    """
    user.password = hash_password(password)
    _datastore.put(user)
    send_password_changed_notice(user)
    password_changed.send(current_app._get_current_object(),
                          user=user)



from flask_wtf import FlaskForm
from wtforms import BooleanField, SubmitField, StringField
from wtforms.validators import DataRequired, ValidationError

from groupbuyorganizer.admin.models import Category


class CreateCategoryForm(FlaskForm):
    category_name = StringField('Category Name', validators=[DataRequired()])
    submit = SubmitField('Submit')

    def validate_category_name(self, category_name):
        category = Category.query.filter_by(name=category_name.data).first()
        if category:
            raise ValidationError('That category name is already being used. Please choose a different one.')


class ApplicationSettingsForm(FlaskForm):
    registration_enabled = BooleanField("Registration Enabled?")
    users_can_see_master_overview = BooleanField("Allow users to see entire order summary?")
    wkhtmltopdf_path = StringField('Absolute file path of wkhtmltopdf binary')
    submit = SubmitField('Submit')


# -*- coding: utf-8 -*-
"""
    flask_security.views
    ~~~~~~~~~~~~~~~~~~~~

    Flask-Security views module

    :copyright: (c) 2012 by Matt Wright.
    :license: MIT, see LICENSE for more details.
"""

from flask import Blueprint, after_this_request, current_app, jsonify, \
    redirect, request
from flask_login import current_user
from werkzeug.datastructures import MultiDict
from werkzeug.local import LocalProxy

from .changeable import change_user_password
from .confirmable import confirm_email_token_status, confirm_user, \
    send_confirmation_instructions
from .decorators import anonymous_user_required, login_required
from .passwordless import login_token_status, send_login_instructions
from .recoverable import reset_password_token_status, \
    send_reset_password_instructions, update_password
from .registerable import register_user
from .utils import url_for_security as url_for
from .utils import config_value, do_flash, get_message, \
    get_post_login_redirect, get_post_logout_redirect, \
    get_post_register_redirect, get_url, login_user, logout_user, \
    slash_url_suffix

# Convenient references
_security = LocalProxy(lambda: current_app.extensions['security'])

_datastore = LocalProxy(lambda: _security.datastore)


def _render_json(form, include_user=True, include_auth_token=False):
    has_errors = len(form.errors) > 0

    if has_errors:
        code = 400
        response = dict(errors=form.errors)
    else:
        code = 200
        response = dict()
        if include_user:
            response['user'] = form.user.get_security_payload()

        if include_auth_token:
            token = form.user.get_auth_token()
            response['user']['authentication_token'] = token

    return jsonify(dict(meta=dict(code=code), response=response)), code


def _commit(response=None):
    _datastore.commit()
    return response


def _ctx(endpoint):
    return _security._run_ctx_processor(endpoint)


@anonymous_user_required
def login():
    """View function for login view"""

    form_class = _security.login_form

    if request.is_json:
        form = form_class(MultiDict(request.get_json()))
    else:
        form = form_class(request.form)

    if form.validate_on_submit():
        login_user(form.user, remember=form.remember.data)
        after_this_request(_commit)

        if not request.is_json:
            return redirect(get_post_login_redirect(form.next.data))

    if request.is_json:
        return _render_json(form, include_auth_token=True)

    return _security.render_template(config_value('LOGIN_USER_TEMPLATE'),
                                     login_user_form=form,
                                     **_ctx('login'))


def logout():
    """View function which handles a logout request."""

    if current_user.is_authenticated:
        logout_user()

    return redirect(get_post_logout_redirect())


@anonymous_user_required
def register():
    """View function which handles a registration request."""

    if _security.confirmable or request.is_json:
        form_class = _security.confirm_register_form
    else:
        form_class = _security.register_form

    if request.is_json:
        form_data = MultiDict(request.get_json())
    else:
        form_data = request.form

    form = form_class(form_data)

    if form.validate_on_submit():
        user = register_user(**form.to_dict())
        form.user = user

        if not _security.confirmable or _security.login_without_confirmation:
            after_this_request(_commit)
            login_user(user)

        if not request.is_json:
            if 'next' in form:
                redirect_url = get_post_register_redirect(form.next.data)
            else:
                redirect_url = get_post_register_redirect()

            return redirect(redirect_url)
        return _render_json(form, include_auth_token=True)

    if request.is_json:
        return _render_json(form)

    return _security.render_template(config_value('REGISTER_USER_TEMPLATE'),
                                     register_user_form=form,
                                     **_ctx('register'))


def send_login():
    """View function that sends login instructions for passwordless login"""

    form_class = _security.passwordless_login_form

    if request.is_json:
        form = form_class(MultiDict(request.get_json()))
    else:
        form = form_class()

    if form.validate_on_submit():
        send_login_instructions(form.user)
        if not request.is_json:
            do_flash(*get_message('LOGIN_EMAIL_SENT', email=form.user.email))

    if request.is_json:
        return _render_json(form)

    return _security.render_template(config_value('SEND_LOGIN_TEMPLATE'),
                                     send_login_form=form,
                                     **_ctx('send_login'))


@anonymous_user_required
def token_login(token):
    """View function that handles passwordless login via a token"""

    expired, invalid, user = login_token_status(token)

    if invalid:
        do_flash(*get_message('INVALID_LOGIN_TOKEN'))
    if expired:
        send_login_instructions(user)
        do_flash(*get_message('LOGIN_EXPIRED', email=user.email,
                              within=_security.login_within))
    if invalid or expired:
        return redirect(url_for('login'))

    login_user(user)
    after_this_request(_commit)
    do_flash(*get_message('PASSWORDLESS_LOGIN_SUCCESSFUL'))

    return redirect(get_post_login_redirect())


def send_confirmation():
    """View function which sends confirmation instructions."""

    form_class = _security.send_confirmation_form

    if request.is_json:
        form = form_class(MultiDict(request.get_json()))
    else:
        form = form_class()

    if form.validate_on_submit():
        send_confirmation_instructions(form.user)
        if not request.is_json:
            do_flash(*get_message('CONFIRMATION_REQUEST',
                     email=form.user.email))

    if request.is_json:
        return _render_json(form)

    return _security.render_template(
        config_value('SEND_CONFIRMATION_TEMPLATE'),
        send_confirmation_form=form,
        **_ctx('send_confirmation')
    )


def confirm_email(token):
    """View function which handles a email confirmation request."""

    expired, invalid, user = confirm_email_token_status(token)

    if not user or invalid:
        invalid = True
        do_flash(*get_message('INVALID_CONFIRMATION_TOKEN'))

    already_confirmed = user is not None and user.confirmed_at is not None

    if expired and not already_confirmed:
        send_confirmation_instructions(user)
        do_flash(*get_message('CONFIRMATION_EXPIRED', email=user.email,
                              within=_security.confirm_email_within))
    if invalid or (expired and not already_confirmed):
        return redirect(get_url(_security.confirm_error_view) or
                        url_for('send_confirmation'))

    if user != current_user:
        logout_user()
        login_user(user)

    if confirm_user(user):
        after_this_request(_commit)
        msg = 'EMAIL_CONFIRMED'
    else:
        msg = 'ALREADY_CONFIRMED'

    do_flash(*get_message(msg))

    return redirect(get_url(_security.post_confirm_view) or
                    get_url(_security.post_login_view))


@anonymous_user_required
def forgot_password():
    """View function that handles a forgotten password request."""

    form_class = _security.forgot_password_form

    if request.is_json:
        form = form_class(MultiDict(request.get_json()))
    else:
        form = form_class()

    if form.validate_on_submit():
        send_reset_password_instructions(form.user)
        if not request.is_json:
            do_flash(*get_message('PASSWORD_RESET_REQUEST',
                     email=form.user.email))

    if request.is_json:
        return _render_json(form, include_user=False)

    return _security.render_template(config_value('FORGOT_PASSWORD_TEMPLATE'),
                                     forgot_password_form=form,
                                     **_ctx('forgot_password'))


@anonymous_user_required
def reset_password(token):
    """View function that handles a reset password request."""

    expired, invalid, user = reset_password_token_status(token)

    if invalid:
        do_flash(*get_message('INVALID_RESET_PASSWORD_TOKEN'))
    if expired:
        send_reset_password_instructions(user)
        do_flash(*get_message('PASSWORD_RESET_EXPIRED', email=user.email,
                              within=_security.reset_password_within))
    if invalid or expired:
        return redirect(url_for('forgot_password'))

    form = _security.reset_password_form()

    if form.validate_on_submit():
        after_this_request(_commit)
        update_password(user, form.password.data)
        do_flash(*get_message('PASSWORD_RESET'))
        login_user(user)
        return redirect(get_url(_security.post_reset_view) or
                        get_url(_security.post_login_view))

    return _security.render_template(
        config_value('RESET_PASSWORD_TEMPLATE'),
        reset_password_form=form,
        reset_password_token=token,
        **_ctx('reset_password')
    )


@login_required
def change_password():
    """View function which handles a change password request."""

    form_class = _security.change_password_form

    if request.is_json:
        form = form_class(MultiDict(request.get_json()))
    else:
        form = form_class()

    if form.validate_on_submit():
        after_this_request(_commit)
        change_user_password(current_user._get_current_object(),
                             form.new_password.data)
        if not request.is_json:
            do_flash(*get_message('PASSWORD_CHANGE'))
            return redirect(get_url(_security.post_change_view) or
                            get_url(_security.post_login_view))

    if request.is_json:
        form.user = current_user
        return _render_json(form)

    return _security.render_template(
        config_value('CHANGE_PASSWORD_TEMPLATE'),
        change_password_form=form,
        **_ctx('change_password')
    )


def create_blueprint(state, import_name):
    """Creates the security extension blueprint"""

    bp = Blueprint(state.blueprint_name, import_name,
                   url_prefix=state.url_prefix,
                   subdomain=state.subdomain,
                   template_folder='templates')

    bp.route(state.logout_url, endpoint='logout')(logout)

    if state.passwordless:
        bp.route(state.login_url,
                 methods=['GET', 'POST'],
                 endpoint='login')(send_login)
        bp.route(state.login_url + slash_url_suffix(state.login_url,
                                                    '<token>'),
                 endpoint='token_login')(token_login)
    else:
        bp.route(state.login_url,
                 methods=['GET', 'POST'],
                 endpoint='login')(login)

    if state.registerable:
        bp.route(state.register_url,
                 methods=['GET', 'POST'],
                 endpoint='register')(register)

    if state.recoverable:
        bp.route(state.reset_url,
                 methods=['GET', 'POST'],
                 endpoint='forgot_password')(forgot_password)
        bp.route(state.reset_url + slash_url_suffix(state.reset_url,
                                                    '<token>'),
                 methods=['GET', 'POST'],
                 endpoint='reset_password')(reset_password)

    if state.changeable:
        bp.route(state.change_url,
                 methods=['GET', 'POST'],
                 endpoint='change_password')(change_password)

    if state.confirmable:
        bp.route(state.confirm_url,
                 methods=['GET', 'POST'],
                 endpoint='send_confirmation')(send_confirmation)
        bp.route(state.confirm_url + slash_url_suffix(state.confirm_url,
                                                      '<token>'),
                 methods=['GET', 'POST'],
                 endpoint='confirm_email')(confirm_email)

    return bp



from setuptools import setup, find_namespace_packages

with open("README.md", "r") as openReadMe:
    long_description = openReadMe.read()

setup(
    name="Group Buy Organizer",
    version="1.0.2",
    author="Mark Michon",
    author_email="markmichon7@gmail.com",
    description="📦 Quickly coordinate large wholesale group orders with this easy to use web app.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/MarkMichon1/Group-Buy-Organizer",
    packages=find_namespace_packages(),
    install_requires=[
        'flask',
        'flask-bcrypt',
        'flask-login',
        'flask-sqlalchemy',
        'flask-wtf',
        'pdfkit',
    ],
    classifiers=[
        "Programming Language :: Python :: 3.7",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)


# -*- coding: utf-8 -*-
"""
    flask_security.registerable
    ~~~~~~~~~~~~~~~~~~~~~~~~~~~

    Flask-Security registerable module

    :copyright: (c) 2012 by Matt Wright.
    :license: MIT, see LICENSE for more details.
"""

from flask import current_app as app
from werkzeug.local import LocalProxy

from .confirmable import generate_confirmation_link
from .signals import user_registered
from .utils import config_value, do_flash, get_message, hash_password, \
    send_mail

# Convenient references
_security = LocalProxy(lambda: app.extensions['security'])

_datastore = LocalProxy(lambda: _security.datastore)


def register_user(**kwargs):
    confirmation_link, token = None, None
    kwargs['password'] = hash_password(kwargs['password'])
    user = _datastore.create_user(**kwargs)
    _datastore.commit()

    if _security.confirmable:
        confirmation_link, token = generate_confirmation_link(user)
        do_flash(*get_message('CONFIRM_REGISTRATION', email=user.email))

    user_registered.send(app._get_current_object(),
                         user=user, confirm_token=token)

    if config_value('SEND_REGISTER_EMAIL'):
        send_mail(config_value('EMAIL_SUBJECT_REGISTER'), user.email,
                  'welcome', user=user, confirmation_link=confirmation_link)

    return user



'''
    flaskext.bcrypt
    ---------------
    
    A Flask extension providing bcrypt hasing and comparison facilities.
    
    :copyright: (c) 2011 by Max Countryman.
    :license: BSD, see LICENSE for more details.
'''

from __future__ import absolute_import
from __future__ import print_function

__version_info__ = ('0', '7', '1')
__version__ = '.'.join(__version_info__)
__author__ = 'Max Countryman'
__license__ = 'BSD'
__copyright__ = '(c) 2011 by Max Countryman'
__all__ = ['Bcrypt', 'check_password_hash', 'generate_password_hash']

from werkzeug.security import safe_str_cmp

try:
    import bcrypt
except ImportError as e:
    print('bcrypt is required to use Flask-Bcrypt')
    raise e

from sys import version_info

PY3 = version_info[0] >= 3


def generate_password_hash(password, rounds=None):
    '''This helper function wraps the eponymous method of :class:`Bcrypt`. It 
    is intended to be used as a helper function at the expense of the 
    configuration variable provided when passing back the app object. In other 
    words this shortcut does not make use of the app object at all.

    To this this function, simple import it from the module and use it in a 
    similar fashion as the method would be used. Here is a quick example::

        from flask.ext.bcrypt import generate_password_hash
        pw_hash = generate_password_hash('hunter2', 10)

    :param password: The password to be hashed.
    :param rounds: The optional number of rounds.
    '''
    return Bcrypt().generate_password_hash(password, rounds)


def check_password_hash(pw_hash, password):
    '''This helper function wraps the eponymous method of :class:`Bcrypt.` It 
    is intended to be used as a helper function at the expense of the 
    configuration variable provided when passing back the app object. In other 
    words this shortcut does not make use of the app object at all.
    
    To this this function, simple import it from the module and use it in a 
    similar fashion as the method would be used. Here is a quick example::
        
        from flask.ext.bcrypt import check_password_hash
        check_password_hash(pw_hash, 'hunter2') # returns True
    
    :param pw_hash: The hash to be compared against.
    :param password: The password to compare.
    '''
    return Bcrypt().check_password_hash(pw_hash, password)


class Bcrypt(object):
    '''Bcrypt class container for password hashing and checking logic using 
    bcrypt, of course. This class may be used to intialize your Flask app 
    object. The purpose is to provide a simple interface for overriding 
    Werkzeug's built-in password hashing utilities.

    Although such methods are not actually overriden, the API is intentionally 
    made similar so that existing applications which make use of the previous 
    hashing functions might be easily adapted to the stronger facility of 
    bcrypt.

    To get started you will wrap your application's app object something like 
    this::

        app = Flask(__name__)
        bcrypt = Bcrypt(app)

    Now the two primary utility methods are exposed via this object, `bcrypt`.
    So in the context of the application, important data, such as passwords, 
    could be hashed using this syntax::

        password = 'hunter2'
        pw_hash = bcrypt.generate_password_hash(password)

    Once hashed, the value is irreversible. However in the case of validating 
    logins a simple hashing of candidate password and subsequent comparison. 
    Importantly a comparison should be done in constant time. This helps 
    prevent timing attacks. A simple utility method is provided for this::

        candidate = 'secret'
        bcrypt.check_password_hash(pw_hash, candidate)

    If both the candidate and the existing password hash are a match 
    `check_password_hash` returns True. Otherwise, it returns False.

    .. admonition:: Namespacing Issues 

        It's worth noting that if you use the format, `bcrypt = Bcrypt(app)` 
        you are effectively overriding the bcrypt module. Though it's unlikely 
        you would need to access the module outside of the scope of the 
        extension be aware that it's overriden.

        Alternatively consider using a different name, such as `flask_bcrypt 
        = Bcrypt(app)` to prevent naming collisions.

    Additionally a configuration value for `BCRYPT_LOG_ROUNDS` may be set in 
    the configuration of the Flask app. If none is provided this will 
    internally be assigned to 12. (This value is used in determining the 
    complexity of the encryption, see bcrypt for more details.)
    
    :param app: The Flask application object. Defaults to None.
    '''

    _log_rounds = 12

    def __init__(self, app=None):
        if app is not None:
            self.init_app(app)

    def init_app(self, app):
        '''Initalizes the application with the extension.

        :param app: The Flask application object.
        '''
        self._log_rounds = app.config.get('BCRYPT_LOG_ROUNDS', 12)
    
    def generate_password_hash(self, password, rounds=None):
        '''Generates a password hash using bcrypt. Specifying `rounds` 
        sets the log_rounds parameter of `bcrypt.gensalt()` which determines 
        the complexity of the salt. 12 is the default value.

        Example usage of :class:`generate_password_hash` might look something 
        like this::

            pw_hash = bcrypt.generate_password_hash('secret', 10)

        :param password: The password to be hashed.
        :param rounds: The optional number of rounds.
        '''

        if not password:
            raise ValueError('Password must be non-empty.')

        if rounds is None:
            rounds = self._log_rounds

        # Python 3 unicode strings must be encoded as bytes before hashing.
        if PY3 and isinstance(password, str):
            password = bytes(password, 'utf-8')

        if not PY3 and isinstance(password, unicode):
            password = password.encode('utf-8')

        return bcrypt.hashpw(password, bcrypt.gensalt(rounds))

    def check_password_hash(self, pw_hash, password):
        '''Tests a password hash against a candidate password. The candidate 
        password is first hashed and then subsequently compared in constant 
        time to the existing hash. This will either return `True` or `False`.

        Example usage of :class:`check_password_hash` would look something 
        like this::

            pw_hash = bcrypt.generate_password_hash('secret', 10)
            bcrypt.check_password_hash(pw_hash, 'secret') # returns True

        :param pw_hash: The hash to be compared against.
        :param password: The password to compare.
        '''

        # Python 3 unicode strings must be encoded as bytes before hashing.
        if PY3 and isinstance(pw_hash, str):
            pw_hash = bytes(pw_hash, 'utf-8')

        if PY3 and isinstance(password, str):
            password = bytes(password, 'utf-8')

        if not PY3 and isinstance(pw_hash, unicode):
            pw_hash = pw_hash.encode('utf-8')

        if not PY3 and isinstance(password, unicode):
            password = password.encode('utf-8')

        return safe_str_cmp(bcrypt.hashpw(password, pw_hash), pw_hash)



from flask import Blueprint, flash, make_response, redirect, render_template, request, url_for
from flask_login import current_user, login_required
import pdfkit

from datetime import timezone

from groupbuyorganizer import database
from groupbuyorganizer.admin.forms import ApplicationSettingsForm, CreateCategoryForm
from groupbuyorganizer.admin.models import Category, Instance, User
from groupbuyorganizer.admin.utilities import admin_protector


admin = Blueprint('admin', __name__)

@admin.route("/category_settings/", methods=['GET', 'POST'])
@login_required
def category_settings():
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    form = CreateCategoryForm()
    if form.validate_on_submit():
        category = Category(name=form.category_name.data)
        database.session.add(category)
        database.session.commit()
        flash('Category successfully added!', 'success')
        return redirect(url_for('admin.category_settings'))
    elif request.method == 'GET':
        categories = Category.query.order_by(Category.name.asc()).all()
    return render_template('category_settings.html', title='Category Settings', categories=categories, form=form)


@admin.route("/category_settings/<int:category_id>/edit/", methods=['GET', 'POST'])
@login_required
def category_edit(category_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    category = Category.query.get_or_404(category_id)
    form = CreateCategoryForm()
    if form.validate_on_submit():
        category.name = form.category_name.data
        database.session.commit()
        flash(f'{category.name} has been edited!', 'info')
        return redirect(url_for('admin.category_settings'))
    elif request.method == 'GET':
        form.category_name.data = category.name
    return render_template('category_edit.html', title='Edit Category Name', form=form)


@admin.route("/categories/<int:category_id>/remove/", methods=['GET'])
@login_required
def category_remove(category_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    category = Category.query.get_or_404(category_id)
    if category.id == 1:
        flash('Access denied', 'danger')
        return redirect(url_for('admin.category_settings'))

    flash_message = 'Category deleted! '
    if category.items:
        flash_message += " Item(s) in this category have been moved to 'Uncategorized.'"
        for item in category.items:
            item.category_id = 1
            database.session.commit()
    database.session.delete(category)
    database.session.commit()
    flash(f'{flash_message}', 'info')
    return redirect(url_for('admin.category_settings'))


@admin.route("/user_settings/")
@login_required
def user_settings():
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    users = User.query.order_by(User.username.asc()).all()
    for user in users: # Making the DateTime database column human readable and converted to local timezone.
        ugly_string = user.date_created.replace(tzinfo=timezone.utc).astimezone(tz=None)
        user.date_created = ugly_string.strftime("%x %X")

    return render_template('user_settings.html', title='User Settings', users=users)


@admin.route("/user_settings/<int:user_id>/promote", methods=['GET'])
@login_required
def promote_user(user_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    user = User.query.get_or_404(user_id)
    admin_protector(user)
    user.is_admin = True
    database.session.commit()
    flash(f'{user.username} has been promoted to admin!', 'info')
    return redirect(url_for('admin.user_settings'))


@admin.route("/user_settings/<int:user_id>/demote")
@login_required
def demote_user(user_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    user = User.query.get_or_404(user_id)
    admin_protector(user)
    user.is_admin = False
    database.session.commit()
    flash(f'{user.username} has been demoted!', 'info')
    return redirect(url_for('admin.user_settings'))


@admin.route("/user_settings/<int:user_id>/disable")
@login_required
def disable_user(user_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    user = User.query.get_or_404(user_id)
    admin_protector(user)
    user.is_admin = False
    user.disabled = True
    database.session.commit()
    flash(f'{user.username} has been disabled!', 'info')
    return redirect(url_for('admin.user_settings'))


@admin.route("/user_settings/<int:user_id>/enable")
@login_required
def enable_user(user_id):
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    user = User.query.get_or_404(user_id)
    admin_protector(user)
    user.disabled = False
    database.session.commit()
    flash(f'{user.username} has been re-enabled!', 'info')
    return redirect(url_for('admin.user_settings'))


@admin.route("/app_settings", methods=['GET', 'POST'])
@login_required
def app_settings():
    if current_user.is_admin == False:
        flash('Access denied', 'danger')
        return redirect(url_for('general.home'))
    instance = Instance.query.first()
    form = ApplicationSettingsForm()
    if form.validate_on_submit():
        instance.registration_enabled = form.registration_enabled.data
        instance.users_can_see_master_overview = form.users_can_see_master_overview.data
        instance.wkhtmltopdf_path = form.wkhtmltopdf_path.data
        database.session.commit()
        flash('Changes saved!', 'info')
        return redirect(url_for('admin.app_settings'))
    elif request.method == 'GET':
        form.registration_enabled.data = instance.registration_enabled
        form.users_can_see_master_overview.data = instance.users_can_see_master_overview
        if instance.wkhtmltopdf_path is not None:
            form.wkhtmltopdf_path.data = instance.wkhtmltopdf_path
    return render_template('app_settings.html', title='Application Settings', form=form)


@admin.route("/app_settings/test_pdf/", methods=['GET'])
@login_required
def test_pdf():
    '''Once you add the absolute file path of the wkhtmltopdf binary in the app settings, this route serves as a way for
    the site admin to check that it is properly configured.
    '''
    instance = Instance.query.first()
    config = pdfkit.configuration(wkhtmltopdf=instance.wkhtmltopdf_path)
    rendered = render_template('test_pdf.html', title='PDF Output Check...')
    pdf = pdfkit.from_string(rendered, False, configuration=config, options={'quiet':''})
    response = make_response(pdf)
    response.headers['Content-Type'] = 'application/pdf'
    response.headers['Content-Disposition'] = 'inline; filename=test_pdf.pdf'
    return response


from flask_wtf import FlaskForm
from wtforms import DecimalField, IntegerField, SelectField, StringField, SubmitField, TextAreaField
from wtforms.validators import InputRequired, DataRequired, ValidationError

from groupbuyorganizer.events.models import Event, Item


class CreateEventForm(FlaskForm):
    event_name = StringField('Event Name', validators=[DataRequired()])
    submit = SubmitField('Submit')

    def validate_event_name(self, event_name):
        event = Event.query.filter_by(name=event_name.data).first()
        if event:
            raise ValidationError('That event name is taken. Please choose a different one.')


class EventExtraChargeForm(FlaskForm):
    extra_charges = DecimalField('Extra Charges')
    event_notes = TextAreaField('Event Notes')
    submit = SubmitField('Update Notes/Charges')


class CreateItemForm(FlaskForm):
    item_name = StringField('Item Name', validators=[DataRequired()])
    price = DecimalField('Price of Item ($)', validators=[DataRequired()])
    packing = IntegerField('Case packing For Item', validators=[DataRequired()])
    category_id = SelectField('Select Item Category', coerce=int, validators=[InputRequired()])
    submit = SubmitField('Submit')


class EditItemForm(FlaskForm):
    item_name = StringField('Item Name', validators=[DataRequired()])
    price = DecimalField('Price of Item ($)', validators=[DataRequired()])
    packing = IntegerField('Case packing For Item', validators=[DataRequired()])
    category_id = SelectField('Select Item Category', coerce=int, validators=[InputRequired()])
    submit = SubmitField('Submit')


class CaseQuantityOrderForm(FlaskForm):
    quantity = SelectField('Cases To Order', coerce=int)
    submit = SubmitField('Update')


class CreateCaseSplitForm(FlaskForm):
    piece_quantity = SelectField('Pieces To Pledge', coerce=int)
    submit = SubmitField('Submit')


class SelectUserFromEventForm(FlaskForm):
    user_to_select = SelectField("Remove User's Transactions From Event", coerce=int)
    submit = SubmitField("Submit")


from flask_login import current_user

from groupbuyorganizer import database
from groupbuyorganizer.admin.models import User
from groupbuyorganizer.events.models import CaseBuy, CasePieceCommit, CaseSplit, Item
from groupbuyorganizer.events.utility_functions import fetch_active_event_items, get_active_participants, \
    get_case_list, how_many_pieces_locked_in, get_cases_reserved_for_item, get_event_total, \
    get_pieces_available_split_item, get_user_total_tuple

'''All repetitive functions used in display objects go in this module.  /events/utilities.py was split into these two
pieces because it was growing a bit messy.'''


class StructuredEventItemList:
    '''This is the main object that jinja pulls data from for the event page.  It organizes items into GroupList items
    by category.
    '''
    def __init__(self, item_list, event_id):
        self.item_list = item_list
        self.event_id = event_id
        self.group_lists = generate_group_lists(self.item_list, item_to_instantiate='event_item',
                                                event_id=self.event_id)


class MyOrderObject:
    def __init__(self, user, item_list, event):
        self.user = user
        self.item_list = item_list
        self.event = event
        self.extra_charges = event.extra_charges
        self.case_list = get_case_list(self.event.id)
        self.event_total = get_event_total(self.case_list)
        self.group_lists = generate_group_lists(item_list=self.item_list, item_to_instantiate='my_order_item',
                                                event_id=self.event.id, user_id=self.user.id, current_user_view=True)
        self.splits = self.get_user_involved_splits()
        self.total_tuple = get_user_total_tuple(self.user, self.event.id, event_total=self.event_total,
                                                event_extra_charges=self.extra_charges)


    def get_user_involved_splits(self):
        split_list = []
        commits = database.session.query(CasePieceCommit).filter(CaseSplit.event_id == self.event.id,
                                                                 CaseSplit.is_complete == True,
                                                                 CasePieceCommit.case_split_id == CaseSplit.id,
                                                                 CasePieceCommit.user_id == self.user.id).all()
        for commit in commits:
            case_split = database.session.query(CaseSplit).filter(CaseSplit.id ==
                                                                                  commit.case_split_id).first()
            if case_split.is_complete == True:
                split_list.append(case_split)

        final_list = []
        reversed_list = sorted(split_list, key=lambda x: x.id, reverse=True)
        for case_split in reversed_list:
            item = database.session.query(Item).filter(Item.id == case_split.item_id).first()
            object_wrapper = CaseSplitItem(case_split, item.packing, self.event.id, item)
            created_by = database.session.query(User.username).filter_by(id=case_split.started_by).first()
            final_list.append((object_wrapper, created_by))
        return final_list


class SummaryObject:
    def __init__(self, event):
        self.event = event
        self.items = fetch_active_event_items(self.event.id)
        self.group_lists = generate_group_lists(self.items, 'summary_item', event_id=self.event.id)
        self.case_list = get_case_list(self.event.id)
        self.total = get_event_total(self.case_list)



class BreakdownObject:
    def __init__(self, event):
        self.event = event
        self.items = fetch_active_event_items(self.event.id)
        self.group_lists = generate_group_lists(self.items, 'user_breakdown_item', event_id=self.event.id)
        self.case_list = get_case_list(self.event.id)
        self.total = get_event_total(self.case_list)


class GroupList:
    '''These are nothing but containers for items with a specific category name.  This helps Jinja loop through lists
    of lists.
    '''
    def __init__(self, category_name, items):
        self.category_name = category_name
        self.items = items
        self.sort()


    def sort(self):
        self.items = sorted(self.items, key=lambda x: x.name)


class EventItem:
    def __init__(self, item, event_id):
        self.item = item
        self.event_id = event_id
        self.name = self.item.name
        self.price = self.item.price
        self.packing = self.item.packing
        self.price = self.item.price
        self.id = self.item.id
        self.case_splits = CaseSplitItemGroup(self.item.case_splits, self.packing, self.event_id)
        self.active_case_splits = self.case_splits.how_many_active_cases()
        self.cases_reserved = get_cases_reserved_for_item(self.item.id, current_user.id)
        self.pieces_locked_in = how_many_pieces_locked_in(self.event_id, self.item.id, current_user.id)
        self.your_total = self.get_total()

    def get_total(self):
        return (self.cases_reserved * self.price) + (self.pieces_locked_in * (self.price / self.packing))


class CaseSplitItemGroup:
    def __init__(self, case_splits, packing, event_id, is_single=False):
        self.case_splits = case_splits
        self.packing = packing
        self.event_id = event_id
        self.is_single = is_single
        self.active_splits = []
        self.complete_splits = []
        self.single_split = None
        self.structure_lists()

    def structure_lists(self):
        reversed_list = sorted(self.case_splits, key=lambda x: x.id, reverse=True)
        for case_split in reversed_list:
            object_wrapper = CaseSplitItem(case_split, self.packing, self.event_id)
            created_by = database.session.query(User.username).filter_by(id=case_split.started_by).first()
            list_tuple = (object_wrapper, created_by)
            if case_split.is_complete == True:
                self.complete_splits.append(list_tuple)
            else:
                self.active_splits.append(list_tuple)
        if self.is_single == True:
            if self.complete_splits:
                self.single_split = self.complete_splits[0]
            else:
                self.single_split = self.active_splits[0]

    def how_many_active_cases(self):
        return len(self.active_splits)


class CaseSplitItem:
    def __init__(self, case_split, item_packing, event_id, item=None):
        self.case_split = case_split
        self.packing = item_packing
        self.event_id = event_id
        self.item = item
        self.name = None
        self.commits = self._get_structured_commits()
        self.is_current_user_involved = self.check_if_current_user_involved()
        self.pieces_available = get_pieces_available_split_item(self.case_split.commits, self.packing)
        if self.item:
            self.name = self.item.name


    def _get_structured_commits(self):
        reversed_list = sorted(self.case_split.commits, key=lambda x: x.id, reverse=True)
        commit_list = []
        for commit in reversed_list:
            commit_id_username = database.session.query(User.id, User.username).filter(CasePieceCommit.event_id ==
                                                                    self.event_id, User.id == commit.user_id).first()
            value = None
            if self.item:
                value = float(self.item.price) * (float(commit.pieces_committed) / float(self.packing))
            commit_list.append((commit_id_username, commit, value))
        return commit_list

    def check_if_current_user_involved(self):
        for commit in self.commits:
            if current_user.id == commit[1].user_id:
                return True
        return False


class UserTotalItem:
    '''This object is used to help present data on the user breakdown page.'''
    def __init__(self, event):
        self.event = event
        self.extra_charges = event.extra_charges
        self.case_list = get_case_list(self.event.id)
        self.event_total = get_event_total(self.case_list)
        self.event_partipicants = get_active_participants(self.event.id, return_length=False)
        self.user_totals_table = self.create_user_table()


    def create_user_table(self):
        table_list = []
        for username in self.event_partipicants:
            item_total_cost = 0
            user = User.query.filter_by(username=username).first()

            # Getting total for case buys
            case_buys = CaseBuy.query.filter_by(user_id=user.id, event_id=self.event.id).all()
            for case_buy in case_buys:
                item_price = database.session.query(Item.price).filter(Item.id == case_buy.item_id).first()[0]
                item_total_cost += float(case_buy.quantity * item_price)

            # Getting total for case splits
            commits = database.session.query(CasePieceCommit).filter(CaseSplit.event_id == self.event.id,
                                                                     CaseSplit.is_complete == True,
                                                                     CasePieceCommit.case_split_id == CaseSplit.id,
                                                                     CasePieceCommit.user_id == user.id).all()
            for commit in commits:
                case_split_item_id = database.session.query(CaseSplit.item_id).filter(CaseSplit.id ==
                                                                                      commit.case_split_id).first()[0]
                item_price_packing = database.session.query(Item.price, Item.packing)\
                                                                        .filter(Item.id == case_split_item_id).first()
                split_total = (float(commit.pieces_committed) / float(item_price_packing[1])) \
                              * float(item_price_packing[0])
                item_total_cost += float(split_total)

            # Extra fee split
            extra_fee_percentage = round(float(item_total_cost)/float(self.event_total), 2)
            extra_fee_split = round(float(self.event.extra_charges) * float(extra_fee_percentage), 2)
            grand_total = round(item_total_cost + extra_fee_split, 2)

            tuple_to_be_appended = (username, item_total_cost, extra_fee_percentage, extra_fee_split, grand_total)
            table_list.append(tuple_to_be_appended)

        return table_list


class MyOrderItem:
    def __init__(self, item, user_id, current_user_view=False):
        self.item = item
        self.user_id = user_id

        self.item_name = self.item.name
        self.name = self.item.name
        self.packing = self.item.packing
        self.case_price = self.item.price
        self.piece_price = round(self.case_price / self.packing, 2)
        self.cases_you_bought = get_cases_reserved_for_item(self.item.id, self.user_id)
        self.case_splits_you_were_in = self.get_total_case_splits_user_in()
        if current_user_view == True:
            self.pieces_reserved = how_many_pieces_locked_in(self.item.event_id, self.item.id, user_id)
        else:
            self.pieces_reserved = how_many_pieces_locked_in(self.item.event_id, self.item.id, current_user.id)
        self.item_total = (self.cases_you_bought * self.item.price) + (self.pieces_reserved * (self.item.price
                                                                                               / self.item.packing))

    def get_total_case_splits_user_in(self):
        count = 0
        case_split_true = CaseSplit.query.filter_by(event_id=self.item.event_id, item_id=self.item.id,
                                                    is_complete=True).all()
        for case_split in case_split_true:
            for commit in case_split.commits:
                if commit.user_id == self.user_id:
                    count += 1
        return count


class SummaryItem:
    def __init__(self, item, event_id):
        self.item = item
        self.event_id = event_id
        self.case_list = get_case_list(self.event_id, item_id=self.item.id)

        self.name = self.item.name
        self.packing = self.item.packing
        self.case_price = self.item.price
        self.piece_price = round(self.case_price / self.packing, 2)
        self.from_case_buy = self.total_cases_bought()
        self.from_case_split = len(self.case_list[1])
        self.cases_bought = self.from_case_buy + self.from_case_split
        self.item_total = self.case_price * self.cases_bought


    def total_cases_bought(self):
        total = 0
        for case_buy in self.case_list[0]:
            total += case_buy.quantity
        return total


class UserBreakdownItem:
    def __init__(self, item, event_id):
        self.item = item
        self.event_id = event_id
        self.case_list = get_case_list(self.event_id, item_id=self.item.id)

        self.name = self.item.name
        self.packing = self.item.packing
        self.case_price = self.item.price
        self.piece_price = round(self.case_price / self.packing, 2)
        self.cases_bought = self.total_cases_bought()
        self.cases_split = len(self.case_list[1])
        self.total_cases = self.cases_bought + self.cases_split

        self.case_buy_table = self.format_case_buy_table()
        self.case_split_cards = []

        for case_split in self.case_list[1]:
            self.case_split_cards.append(CaseSplitItem(case_split, self.item.packing, self.event_id, item=self.item))


    def total_cases_bought(self):
        total = 0
        for case_buy in self.case_list[0]:
            total += case_buy.quantity
        return total


    def format_case_buy_table(self):
        table = []
        for case_buy in self.case_list[0]:
            username = database.session.query(User.username).filter(User.id == case_buy.user_id).first()[0]
            table.append((username, case_buy.quantity))
        table.sort()
        return table


def generate_group_lists(item_list, item_to_instantiate, event_id=None, user_id=None, current_user_view=False):
    '''Organizing items by category is a common task in this library.  This will organize items in different categories
    into their own groups, and will instantiate these values into a variety of items (sub-components to display item or
    purchase data on the page).
    '''

    group_lists = []
    current_category = ''
    categorized_items = []
    for group in item_list:
        if group[1] != current_category: #reset
            if categorized_items:
                group_lists.append(GroupList(current_category, categorized_items))
            current_category = group[1]
            categorized_items = []
        if item_to_instantiate == 'event_item':
            categorized_items.append(EventItem(group[0], event_id))
        elif item_to_instantiate == 'my_order_item':
            categorized_items.append(MyOrderItem(group[0], user_id, current_user_view=current_user_view))
        elif item_to_instantiate == 'summary_item':
            categorized_items.append(SummaryItem(group[0], event_id))
        elif item_to_instantiate == 'user_breakdown_item':
            categorized_items.append(UserBreakdownItem(group[0], event_id))
    group_lists.append(GroupList(current_category, categorized_items))
    return group_lists


from flask import Blueprint, render_template

errors = Blueprint('errors', __name__)


@errors.app_errorhandler(404)
def error_404(error):
    return render_template('404.html', title='404'), 404

@errors.app_errorhandler(500)
def error_404(error):
    return render_template('500.html', title='500'), 500


from groupbuyorganizer import database
from groupbuyorganizer.admin.models import Category, User
from groupbuyorganizer.events.models import CaseBuy, CasePieceCommit, CaseSplit, Item

'''All repetitive functions used in display objects go in this module.  /events/utilities.py was split into these two
pieces because it was growing a bit messy.'''

def get_pieces_available_split_item(case_split, packing):
    '''This iterates over open case splits, and based off of the total commits, will return a remainder of how many
    items are left.  This is used in the CaseSplitItem class, and in the event route to properly render forms.
    '''

    pieces_left = packing

    for commit in case_split:
        pieces_left -= commit.pieces_committed
    return pieces_left


def return_qty_price_select_field(max_pieces, item_price, item_packing, whole_cases=False):
    '''Taking in the item price, item packing, and maximum allowable pieces, this will return a nicely formatted list
    for flask-wtf's SelectField, which displays the total piece sum next to the piece count.'''

    choices_list = []
    if whole_cases == False:
        for i in range(max_pieces):
            choices_list.append((i + 1, f'{i + 1}/{item_packing} - ${round((i + 1) * (item_price / item_packing), 2)}'))
    else:
        for i in range(max_pieces):
            choices_list.append((i , f'{i} - ${round(i * item_price, 2)}'))
    return choices_list


def get_case_list(event_id, item_id=None):
    '''This query is used several times, so this function exists to reduce redundant code.  It returns a list of both
    case buys, and completed case splits.
    '''

    if item_id == None:
        case_buys = CaseBuy.query.filter_by(event_id=event_id).all()
        case_splits = CaseSplit.query.filter_by(event_id=event_id, is_complete=True).all()
    else:
        case_buys = CaseBuy.query.filter_by(event_id=event_id, item_id=item_id).all()
        case_splits = CaseSplit.query.filter_by(event_id=event_id, item_id=item_id, is_complete=True).all()
    return case_buys, case_splits


def get_event_total(case_list):
    total_cost = 0
    for case_order in case_list[0]:
        item = Item.query.filter_by(id=case_order.item_id).first()
        total_cost += (case_order.quantity * item.price)

    for case_split in case_list[1]:
        if case_split.is_complete == True:
            item = Item.query.filter_by(id=case_split.item_id).first()
            total_cost += item.price
    return total_cost


def get_active_participants(event_id, return_length):
    '''This gets a list of usernames who have partipicant in a particular event.  Depending on what you want the
    function to do, it can either return the length of the list, or the list by itself.
    '''

    case_order_user_ids = database.session.query(User.username).filter(CaseBuy.event_id == event_id,
                                                                 User.id == CaseBuy.user_id).all()
    case_split_user_ids = database.session.query(User.username).filter(CasePieceCommit.event_id == event_id,
                                                                 CasePieceCommit.user_id == User.id).all()

    total_user_set = set()
    for user in case_order_user_ids:
        total_user_set.add(user)
    for user in case_split_user_ids:
        total_user_set.add(user)

    if return_length == True:
        return len(total_user_set)
    else:
        untupled_list = []
        to_list = list(total_user_set)
        for user_name in to_list:
            untupled_list.append(user_name[0])
        untupled_list.sort()
        return untupled_list

def is_user_active(user_id, event_id):
    '''This returns a boolean value on whether this particular user ID is active in the group buy event or not.'''

    case_buys = CaseBuy.query.filter_by(user_id=user_id, event_id=event_id).first()
    split_commits = CasePieceCommit.query.filter_by(user_id=user_id, event_id=event_id).first()
    if case_buys or split_commits:
        return True

    return False


def is_event_active(event_id):
    '''This returns a boolean value on whether this particular event is active or not.'''

    case_buys = CaseBuy.query.filter_by(event_id=event_id).first()
    split_commits = CasePieceCommit.query.filter_by(event_id=event_id).first()
    if case_buys or split_commits:
        return True

    return False


def fetch_user_items(user_id, event_id):
    '''Returns a list of items that the '''
    item_set = set()

    # Getting case buys
    case_buy_items = database.session.query(CaseBuy) \
        .filter(CaseBuy.user_id == user_id, CaseBuy.event_id == event_id).all()

    # Getting case splits
    split_commit_items = CasePieceCommit.query.filter_by(user_id=user_id, event_id=event_id).all()
    filtered_commits = []
    for commit in split_commit_items:
        case_split = CaseSplit.query.filter_by(id=commit.case_split_id).first()
        if case_split.is_complete == True:
            filtered_commits.append((case_split.item_id,))

    # Extracting items and category names out of case buys
    for case_buy in case_buy_items:
        item = Item.query.filter_by(id=case_buy.item_id).first()
        category_name = database.session.query(Category.name).filter_by(id=item.category_id).first()
        item_set.add((item, category_name[0]))

    # Extracting items and category names out of case split commits
    for item_id in filtered_commits:
        item = Item.query.filter_by(id=item_id[0]).first()
        category_name = database.session.query(Category.name).filter_by(id=item.category_id).first()
        item_set.add((item, category_name[0]))

    # Final formatting
    to_list = list(item_set)
    to_list = sorted(to_list, key=lambda x: x[0].name)
    return to_list

def fetch_active_event_items(event_id):
    '''Taking in the event ID, it returns all items that were involved in case buys or completed case splits in the
    given event.
    '''

    item_set = set()
    case_buy_items = database.session.query(Item).filter(CaseBuy.event_id == event_id, CaseBuy.item_id == Item.id).all()
    case_split_items = database.session.query(Item).filter(CaseSplit.event_id == event_id,
                                                           CaseSplit.item_id == Item.id,
                                                           CaseSplit.is_complete == True).all()
    for item in case_buy_items:
        item_set.add(item)
    for item in case_split_items:
        item_set.add(item)
    item_category_pair = []
    for item in item_set:
        category_name = database.session.query(Category.name).filter(Category.id == item.category_id).first()[0]
        item_category_pair.append((item, category_name))
    sorted_list = sorted(item_category_pair, key=lambda x: x[1])
    return sorted_list


def get_cases_reserved_for_item(item_id, user_id):
    '''This will return how many cases a user has purchased of a given item, otherwise it returns zero.'''

    cases_reserved = CaseBuy.query.filter_by(item_id=item_id, user_id=user_id).first()
    if cases_reserved is None:
        return 0
    return cases_reserved.quantity

def generate_active_user_select_field(event_id):

    case_buy_users = database.session.query(User.id, User.username).filter(CaseBuy.event_id == event_id, User.id == CaseBuy.user_id).all()
    case_split_users = database.session.query(User.id, User.username).filter(CaseSplit.event_id == event_id, CasePieceCommit.case_split_id == CaseSplit.id,
                        User.id == CasePieceCommit.user_id).all()
    case_buy_set = set(case_buy_users)
    case_split_set = set(case_split_users)
    total_user_set = set()
    for user in case_buy_set:
        total_user_set.add(user)
    for user in case_split_set:
        total_user_set.add(user)
    total_event_users = list(total_user_set)
    total_event_users.sort(key=lambda x: x[1])
    return total_event_users


def how_many_pieces_locked_in(event_id, item_id, user_id):
    '''This will return how many case pieces a given user has locked in (aka, the case split is "complete").'''

    count = 0
    splits = CaseSplit.query.filter(CaseSplit.is_complete == True, CaseSplit.event_id == event_id,
                                    CaseSplit.item_id == item_id).all()
    for split in splits:
        commits = CasePieceCommit.query.filter(CasePieceCommit.event_id == event_id,
                                               CasePieceCommit.case_split_id == split.id, CasePieceCommit.user_id ==
                                               user_id).all()
        for commit in commits:
            count += commit.pieces_committed
    return count


def get_user_total_tuple(user, event_id, event_total=0, event_extra_charges=0):
    item_total_cost = 0

    # Getting total for case buys
    case_buys = CaseBuy.query.filter_by(user_id=user.id, event_id=event_id).all()
    for case_buy in case_buys:
        item_price = database.session.query(Item.price).filter(Item.id == case_buy.item_id).first()[0]
        item_total_cost += case_buy.quantity * item_price

    # Getting total for case splits
    commits = database.session.query(CasePieceCommit).filter(CaseSplit.event_id == event_id,
                                                             CaseSplit.is_complete == True,
                                                             CasePieceCommit.case_split_id == CaseSplit.id,
                                                             CasePieceCommit.user_id == user.id).all()
    for commit in commits:
        case_split_item_id = database.session.query(CaseSplit.item_id).filter(CaseSplit.id ==
                                                                              commit.case_split_id).first()[0]
        item_price_packing = database.session.query(Item.price, Item.packing) \
            .filter(Item.id == case_split_item_id).first()
        split_total = (item_price_packing[0] / item_price_packing[1]) * commit.pieces_committed
        item_total_cost += split_total

    # Extra fee split
    extra_fee_percentage = round(item_total_cost / event_total, 2)
    extra_fee_split = round(event_extra_charges * extra_fee_percentage, 2)
    grand_total = round(item_total_cost + extra_fee_split, 2)

    user_total_tuple = (user.username, item_total_cost, extra_fee_percentage, extra_fee_split, grand_total)
    return user_total_tuple